/*
 * ExactSubtreeMatch.cpp
 *
 *  Created on: 9. 2. 2014
 *      Author: Jan Travnicek
 */

#include "ExactSubtreeMatch.h"
#include <exception/AlibException.h>
#include <tree/ranked/RankedTree.h>
#include <tree/ranked/PrefixRankedNotation.h>
#include <tree/unranked/UnrankedTree.h>

#include <deque>
#include <foreach>

namespace arbology {

namespace exact {

std::set<unsigned> ExactSubtreeMatch::match(const tree::Tree& subject, const tree::Tree& pattern) {
	std::set<unsigned> data;
	Accept((void*) &data, subject.getData(), pattern.getData(), ExactSubtreeMatch::EXACT_SUBTREE_MATCH);
	return data;
}

bool ExactSubtreeMatch::matchHelper(const tree::UnrankedNode& subject, const tree::UnrankedNode& pattern) {
	if(subject.getSymbol() != pattern.getSymbol()) return false;
	if(subject.getChildren().size() != pattern.getChildren().size()) return false;
	for(const std::tuple<const tree::UnrankedNode*, const tree::UnrankedNode*>& childs : std::make_pair_foreach(subject.getChildren(), pattern.getChildren())) {
		if(!matchHelper(*std::get<0>(childs), *std::get<1>(childs))) return false;
	}
	return true;
}

bool ExactSubtreeMatch::matchHelper(const tree::RankedNode& subject, const tree::RankedNode& pattern) {
	if(subject.getSymbol() != pattern.getSymbol()) return false;
	// ranked symbols are the same test for number of children is not needed
	for(const std::tuple<const tree::RankedNode*, const tree::RankedNode*>& childs : std::make_pair_foreach(subject.getChildren(), pattern.getChildren())) {
		if(!matchHelper(*std::get<0>(childs), *std::get<1>(childs))) return false;
	}
	return true;
}

void ExactSubtreeMatch::matchInternal(unsigned& index, std::set<unsigned>& occ, const tree::UnrankedNode& subject, const tree::UnrankedNode& pattern) {
	if(matchHelper(subject, pattern)) occ.insert(index);
	index++;
	for(const tree::UnrankedNode* child : subject.getChildren()) {
		matchInternal(index, occ, *child, pattern);
	}
}

void ExactSubtreeMatch::matchInternal(unsigned& index, std::set<unsigned>& occ, const tree::RankedNode& subject, const tree::RankedNode& pattern) {
	if(matchHelper(subject, pattern)) occ.insert(index);
	index++;
	for(const tree::RankedNode* child : subject.getChildren()) {
		matchInternal(index, occ, *child, pattern);
	}
}

std::set<unsigned> ExactSubtreeMatch::match(const tree::UnrankedTree& subject, const tree::UnrankedTree& pattern) {
	unsigned i = 0;
	std::set<unsigned> occ;
	matchInternal(i, occ, subject.getRoot(), pattern.getRoot());
	return occ;
}

std::set<unsigned> ExactSubtreeMatch::match(const tree::RankedTree& subject, const tree::RankedTree& pattern) {
	unsigned i = 0;
	std::set<unsigned> occ;
	matchInternal(i, occ, subject.getRoot(), pattern.getRoot());
	return occ;
}

std::set<unsigned> ExactSubtreeMatch::match(const tree::PrefixRankedNotation& subject, const tree::PrefixRankedNotation& pattern) {
	std::set<unsigned> occ;
	for(unsigned i = 0; i <= subject.getContent().size() - pattern.getContent().size(); i++) {
		unsigned j = 0;
		for(; j < pattern.getContent().size(); j++) {
			if(pattern.getContent()[j] != subject.getContent()[i+j]) break;
		}

		if( j == pattern.getContent().size() )
			occ.insert(i);
	}
	return occ;
}

void ExactSubtreeMatch::Visit(void* data, const tree::UnrankedTree& subject, const tree::UnrankedTree& pattern) const {
	std::set<unsigned> & res = *((std::set<unsigned>*) data);
	res = this->match(subject, pattern);
}

void ExactSubtreeMatch::Visit(void* data, const tree::RankedTree& subject, const tree::RankedTree& pattern) const {
	std::set<unsigned> & res = *((std::set<unsigned>*) data);
	res = this->match(subject, pattern);
}

void ExactSubtreeMatch::Visit(void*, const tree::RankedPattern&, const tree::RankedPattern&) const {
	throw exception::AlibException("Unsupported tree type RankedPattern");
}

void ExactSubtreeMatch::Visit(void* data, const tree::PrefixRankedNotation& subject, const tree::PrefixRankedNotation& pattern) const {
	std::set<unsigned> & res = *((std::set<unsigned>*) data);
	res = this->match(subject, pattern);
}

void ExactSubtreeMatch::Visit(void*, const tree::UnrankedPattern&, const tree::UnrankedPattern&) const {
	throw exception::AlibException("Unsupported tree type UnrankedPattern");
}

const ExactSubtreeMatch ExactSubtreeMatch::EXACT_SUBTREE_MATCH;

} /* namespace exact */

} /* namespace arbology */
